/*
**      types.c - Ember type handling routines
**
**
** Copyright (c) 2001  Hughes Technologies Pty Ltd
**
** Permission to use, copy, and distribute for non-commercial purposes,
** is hereby granted without fee, providing that the above copyright
** notice appear in all copies and that both the copyright notice and this
** permission notice appear in supporting documentation.
**
** This software is provided "as is" without any expressed or implied warranty.
**
**
*/

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <ctype.h>
#include <sys/types.h>

#include "config/config.h"
#include "regexp/regexp.h"
#include "ember.h"
#include "parser.h"

#ifdef HAVE_STRINGS_H
#include <strings.h>
#endif



/************************************************************************
** Predeclaration of type functions
*/

int	intPack();
char	*intUnpack();
int	intCompare();
int	intMath();
int	intCheck();

int	uintPack();
char	*uintUnpack();
int	uintCompare();
int	uintMath();

int	textPack();
char	*textUnpack();
int	textCompare();
int	textMath();
int	textCheck();

int	realPack();
char	*realUnpack();
int	realCompare();
int	realMath();
int	realCheck();

int	moneyPack();
char	*moneyUnpack();

/************************************************************************
** Type table
**
** NOTE : The first 3 entires in the table _MUST_NOT_ be modified.  If
** the order of the first 3 entries is changed then things will break.
** The order must be 'real', 'int', 'char'.  If you wish to add your
** own types they must be defined _after_ the definition of 'char'
*/


#define	TYPE_REAL_OFF	0
#define	TYPE_INT_OFF	1
#define	TYPE_TEXT_OFF	2
#define	TYPE_UINT_OFF	3
#define	TYPE_MONEY_OFF	4
#define TYPE_INDEX	3

static e_type	types[] = {
	{ "real", 8, realPack, realUnpack, realCompare, realMath, realCheck},
	{ "int", 4, intPack, intUnpack, intCompare, intMath, intCheck},
	{ "text", -1, textPack, textUnpack, textCompare, textMath, textCheck},
	{ "uint", 4, uintPack, uintUnpack, uintCompare, uintMath, intCheck},
	{ "money", 8, moneyPack, moneyUnpack, realCompare, realMath, realCheck},
	{ "", 0, NULL, NULL, NULL, NULL, NULL }
};




/************************************************************************
** Type definition functions
*/


static	void _debugTypeTrap()
{
}


int eTypeDetermineType(val)
	char	*val;
{
	int	index;

	index = 0;
	while(types[index].name)
	{
		if ( (*types[index].check)(val) == 0)
			return(index);
		index++;
	}
	return(-1);
}




/************************************************************************
** Integer functions
*/



static int my_atoi(val, error)
	char	*val;
	int	*error;
{
	int  	res,
		neg;
	char	*cp;

	*error = 0;
	neg = res = 0;
	cp = val;
	if (*cp == '-')
	{
		neg++;
		cp++;
	}
	while(isdigit((int)*cp))
	{
		res *= 10;
		res += *cp - '0';
		cp++;
	}
	if (*cp != 0)
	{
		*error = 1;
		if (*cp == ' ' && *(cp+1) == 0) *error = 0;
		if (*cp == '\n' && *(cp+1) == 0) *error = 0;
		if (*error == 1)
			return(-1);
	}
	if (neg)
		res = 0 - res;
	return(res);
}


#ifdef NOTDEF
static int oct2dec(val, error)
        char    *val;
	int	*error;
{
        int     decval,
                octval;
        char    *cd;

	*error = 0;
        cd = val + 1;
        decval = 0;
        while(*cd)
        {
		if (!isdigit((int)*cd))
		{
			*error = 1;
                        return(-1);
                }
                decval *= 8;
                octval = *cd - '0';
                if (octval > 7)
                {
			/* 
			** Bad octal format
			*/
			*error = 1;
                        return(-1);
                }
                decval += octval;
                cd++;
        }
        return(decval);
}

#endif /* NOTDEF */


static int hex2dec(val, error)
        char    *val;
	int	*error;
{
        int     decval;
        char    *cd;

        cd = val + 2;
        decval = 0;
        while(*cd)
        {
		if (!isdigit((int)*cd) && tolower(*cd) < 'a' 
			&& tolower(*cd) > 'f')
		{
			*error = 1;
                        return(-1);
		}
                decval *= 16;
                if (*cd >= 'a')
                        decval += *cd - 'a' + 10;
                else
                if (*cd >= 'A')
                        decval += *cd - 'A' + 10;
                else
                        decval += *cd - '0';
                cd++;
        }
        return(decval);
}



int intPack(textRep, value)
	char	*textRep;
	e_value	*value;
{
	int	error;
	static	e_type	*typePtr;

	if (typePtr == NULL)
		typePtr = eTypeGetType(EMBER_TYPE_INT);
	/*
	** Do base conversions if it looks like it's needed
	*/

	if (*textRep == '0' && *(textRep+1) == 'x')
	{
		value->value.intVal = hex2dec(textRep,&error);
	}
	else
	{
		value->value.intVal = my_atoi(textRep,&error);
	}
	value->type = EMBER_TYPE_INT;
	value->typePtr = typePtr;
	value->length = 4;
	if (error)
		return(-1);
	return(4);
}



char *intUnpack(value)
	e_value	*value;
{
	static	char valBuf[20];

	sprintf(valBuf,"%d",value->value.intVal);
	return(valBuf);
}

int uintPack(textRep, value)
	char	*textRep;
	e_value	*value;
{
	int	length;
        static  e_type  *typePtr;

	if (typePtr == NULL)
		typePtr = eTypeGetType(EMBER_TYPE_UINT);
	length = intPack(textRep, value);
	value->type = EMBER_TYPE_UINT;
	value->typePtr = typePtr;
	return(length);
}


char *uintUnpack(value)
	e_value	*value;
{
	static	char valBuf[20];

	sprintf(valBuf,"%u",value->value.uintVal);
	return(valBuf);
}



int intCompare(script,v1, v2, op, res)
	ember	*script;
	e_value	*v1,
		*v2;
	int	op,
		*res;
{
	int	int1 = v1->value.intVal,
		int2 = v2->value.intVal;

	switch(op)
	{
		case COND_GE:
			*res = (int1 >= int2);
			break;
		case COND_LE:
			*res = (int1 <= int2);
			break;
		case COND_NE:
			*res = (int1 != int2);
			break;
		case COND_EQ:
			*res = (int1 == int2);
			break;
		case COND_GT:
			*res = (int1 > int2);
			break;
		case COND_LT:
			*res = (int1 < int2);
			break;
		case COND_RE:
			return(-1);
			break;
	}
	return(0);
}



int intMath(script,v1,v2,op, result)
	ember	*script;
	e_value	*v1,
		*v2,
		*result;
	int	op;
{
	int	int1,
		int2;

	result->type = v1->type;
	result->typePtr = v1->typePtr;
	result->length = v1->length;

	int1 = v1->value.intVal;
	int2 = v2->value.intVal;

	switch(op)
	{
		case OP_ADD:
			result->value.intVal = int1 + int2;
			break;
		case OP_SUB:
			result->value.intVal = int1 - int2;
			break;
		case OP_MUL:
			result->value.intVal = int1 * int2;
			break;
		case OP_DIV:
			if (int2 == 0)
			{
				strcpy(script->errorMsg, "Divide by zero");
				return(-1);
			}
			result->value.intVal = int1 / int2;
			break;
		case OP_AND:
			result->value.intVal = int1 & int2;
			break;
		case OP_OR:
			result->value.intVal = int1 | int2;
			break;
		case OP_MOD:
			if (int2 == 0)
			{
				strcpy(script->errorMsg, "Divide by zero");
				exit(1);
			}
			result->value.intVal = int1 % int2;
			break;
	}
	return(0);
}


int uintCompare(script,v1, v2, op, res)
	ember	*script;
	e_value	*v1,
		*v2;
	int	op,
		*res;
{
	u_int	int1 = v1->value.uintVal,
		int2 = v2->value.uintVal;

	switch(op)
	{
		case COND_GE:
			*res = (int1 >= int2);
			break;
		case COND_LE:
			*res = (int1 <= int2);
			break;
		case COND_NE:
			*res = (int1 != int2);
			break;
		case COND_EQ:
			*res = (int1 == int2);
			break;
		case COND_GT:
			*res = (int1 > int2);
			break;
		case COND_LT:
			*res = (int1 < int2);
			break;
		case COND_RE:
			return(-1);
			break;
	}
	return(0);
}



int uintMath(script,v1,v2,op, result)
	ember	*script;
	e_value	*v1,
		*v2,
		*result;
	int	op;
{
	u_int	val1,
		val2;

	result->type = v1->type;
	result->typePtr = v1->typePtr;
	result->length = v1->length;

	val1 = v1->value.uintVal;
	val2 = v2->value.uintVal;

	switch(op)
	{
		case OP_ADD:
			result->value.uintVal = val1+val2;
			break;
		case OP_SUB:
			result->value.uintVal = val1-val2;
			break;
		case OP_MUL:
			result->value.uintVal = val1*val2;
			break;
		case OP_DIV:
			if (val2 == 0)
			{
				/* FIX
                		runError("Divide by zero");
				*/
				exit(1);
			}
			result->value.uintVal = val1/val2;
			break;
		case OP_AND:
			result->value.uintVal = val1&val2;
			break;
		case OP_OR:
			result->value.uintVal = val1|val2;
			break;
		case OP_MOD:
			if (val2 == 0)
			{
				/* FIX
                		runError("Divide by zero");
				*/
				exit(1);
			}
			result->value.uintVal = val1%val2;
			break;
	}
	return(0);
}



int intCheck(str)
	char	*str;
{
	int	val;
	char	*cp;

	cp = str;

	/*	
	** First thing, check if it's got a dot.  If it's got a dot then
	** it was rejected by realCheck.  Probably an IP address
	*/
	if ((char *)index(str,'.') != NULL)
	{
		return(-1);
	}
	if (*cp == '-')
		cp++;
	
	/* Hex number */
	if (*cp == '0' && *(cp+1) == 'x')
	{
		cp += 2;
		while (*cp)
		{
			if ( isdigit((int)*cp) || 
			     ( *cp >= 'A' && *cp <= 'F') ||
			     ( *cp >= 'a' && *cp <= 'A'))
			{
				cp++;
				continue;
			}	
			return(-1);
		}
	}
	
	if (sscanf(cp,"%d",&val) == 1)
	{
		return(0);
	}
	return(-1);
}



int uintCheck(str)
	char	*str;
{
	int	val;
	char	*cp;

	cp = str;

	/*	
	** First thing, check if it's got a dot.  If it's got a dot then
	** it was rejected by realCheck.  Probably an IP address
	*/
	if ((char *)index(str,'.') != NULL)
	{
		return(-1);
	}
	
	/* Hex number */
	if (*cp == '0' && *(cp+1) == 'x')
	{
		cp += 2;
		while (*cp)
		{
			if ( isdigit((int)*cp) || 
			     ( *cp >= 'A' && *cp <= 'F') ||
			     ( *cp >= 'a' && *cp <= 'A'))
			{
				cp++;
				continue;
			}	
			return(-1);
		}
	}
	
	if (sscanf(cp,"%d",&val) == 1)
	{
		return(0);
	}
	return(-1);
}


/************************************************************************
** Real functions
*/

int moneyPack(textRep, value)
        char    *textRep;
	e_value	*value;
{
        double  packVal;
	static	e_type	*typePtr;

	if (typePtr == NULL)
		typePtr = eTypeGetType(EMBER_TYPE_MONEY);

        /*
        ** Ensure it's a reasonable value
        */
        if (sscanf(textRep,"%lf",&packVal) != 1)
        {
                return(-1);
        }

	value->value.realVal = packVal;
	value->type = EMBER_TYPE_MONEY;
	value->typePtr = typePtr;
	value->length = 8;
	return(8);
}


char *moneyUnpack(value)
	e_value	*value;
{
        static	char valBuf[50];
          
        sprintf(valBuf,"%.2f",value->value.realVal);
        return(valBuf);
}
          


/************************************************************************
** Real functions
*/

int realPack(textRep, value)
	char	*textRep;
	e_value	*value;
{
	double	packVal;
	static	e_type	*typePtr;

	if (typePtr == NULL)
		typePtr = eTypeGetType(EMBER_TYPE_REAL);

	/*
	** Ensure it's a reasonable value
	*/
	if (sscanf(textRep,"%lf",&packVal) != 1)
	{
		return(-1);
	}

	value->value.realVal = packVal;
	value->type = EMBER_TYPE_REAL;
	value->typePtr = typePtr;
	value->length = 8;
	return(8);
}



char *realUnpack(value)
	e_value	*value;
{
	static char valBuf[50];

	sprintf(valBuf,"%.16g",value->value.realVal);
	return(valBuf);
}




int realCompare(script,v1, v2, op, res)
	ember	*script;
	e_value	*v1,
		*v2;
	int	op,
		*res;
{
	double	real1 = v1->value.realVal,
		real2 = v2->value.realVal;

	switch(op)
	{
		case COND_GE:
			*res = (real1 >= real2);
			break;
		case COND_LE:
			*res = (real1 <= real2);
			break;
		case COND_NE:
			*res = (real1 != real2);
			break;
		case COND_EQ:
			*res = (real1 == real2);
			break;
		case COND_GT:
			*res = (real1 > real2);
			break;
		case COND_LT:
			*res = (real1 < real2);
			break;
		case COND_RE:
			return(-1);
			break;
	}
	return(0);
}




int realMath(script,v1,v2,op, result)
	ember	*script;
	e_value	*v1,
		*v2,
		*result;
	int	op;
{
	double	val1,
		val2;

	result->type = v1->type;
	result->typePtr = v1->typePtr;
	result->length = v1->length;

	val1 = v1->value.realVal;
	val2 = v2->value.realVal;


	switch(op)
	{
		case OP_ADD:
			result->value.realVal = val1+val2;
			break;
		case OP_SUB:
			result->value.realVal = val1-val2;
			break;
		case OP_MUL:
			result->value.realVal = val1*val2;
			break;
		case OP_DIV:
			if (val2 == 0)
			{
				/* FIX
                		runError("Divide by zero");
				*/
				exit(1);
			}
			result->value.realVal = val1/val2;
			break;
		case OP_AND:
		case OP_OR:
		case OP_MOD:
			return(-1);
			break;
	}
	return(0);
}




int realCheck(str)
	char	*str;
{
	double	val;
	char	*cp;

	/*
	** If there's no '.' the we'll treat it as an integer value
	*/
	cp = (char *)index(str,'.');
	if (!cp)
	{
		return(-1);
	}
	if ((char *)index(cp+1,'.') != NULL)
	{
		/* multiple dots.  Probably an IP address */
		return(-1);
	}
	
	if (sscanf(str,"%lf",&val) == 1)
	{
		return(0);
	}
	return(-1);
}



/************************************************************************
** Char functions
*/

int textPack(textRep, value)
	char	*textRep;
	e_value	*value;
{
	int	tmpLen;
	static	e_type	*typePtr;

	if (typePtr == NULL)
		typePtr = eTypeGetType(EMBER_TYPE_TEXT);

	tmpLen = strlen(textRep);
	value->value.textVal = strdup(textRep);
	value->type = EMBER_TYPE_TEXT;
	value->typePtr = typePtr;
	value->length = tmpLen + 1;
	return(tmpLen);
}


char *textUnpack(value)
	e_value	*value;
{
	return(value->value.textVal);
}



void regerror(msg)
	char	*msg;
{
	char	buf[160];

	sprintf(buf,"Regular expression error - %s",msg);
	/* FIX
	runError(buf);
	*/
	exit(1);
}


int textCompare(script,v1, v2, op, res)
	ember	*script;
	e_value	*v1,
		*v2;
	int	op,
		*res;
{
	char	*text1 = v1->value.textVal,
		*text2 = v2->value.textVal;

	switch(op)
	{
		case COND_GE:
			*res = (strcmp(text1,text2) >= 0);
			break;
		case COND_LE:
			*res = (strcmp(text1,text2) <= 0);
			break;
		case COND_NE:
			*res = (strcmp(text1,text2) != 0);
			break;
		case COND_EQ:
			*res = (strcmp(text1,text2) == 0);
			break;
		case COND_GT:
			*res = (strcmp(text1,text2) > 0);
			break;
		case COND_LT:
			*res = (strcmp(text1,text2) < 0);
			break;
		case COND_RE:
		{
			regexp	*reg;
			reg = regcomp(text2);
			*res = regexec(reg,text1);
			(void)free(reg);
			break;
		}
	}
	return(0);
}


int textMath(script,v1,v2,op, result)
	ember	*script;
	e_value	*v1,
		*v2,
		*result;
	int	op;
{
	static	char	intBuffer[100];
	char 	*t1 = v1->value.textVal,
		*t2 = v2->value.textVal,
		*buf;
	int	len;

	if (op != OP_ADD)
	{
		return(-1);
	}
	len = strlen(t1)+strlen(t2);
	if (len > 99)
	{
		buf = malloc(len+1);
	}
	else
	{
		buf = intBuffer;
	}
	strcpy(buf,t1);
	strcat(buf,t2);
	eSymSetTextValue(script, result, buf);
	if (buf != intBuffer)
	{
		free(buf);
	}
	return(0);
}

int textCheck(str)
	char	*str;
{
	/* Anything matches the char type */
	return(0);
}


e_type *eTypeGetType(type)
	int	type;
{
	switch(type)
	{
		case EMBER_TYPE_INT:
			return(&types[TYPE_INT_OFF]);
		case EMBER_TYPE_UINT:
			return(&types[TYPE_UINT_OFF]);
		case EMBER_TYPE_REAL:
			return(&types[TYPE_REAL_OFF]);
		case EMBER_TYPE_TEXT:
			return(&types[TYPE_TEXT_OFF]);
		case EMBER_TYPE_MONEY:
			return(&types[TYPE_MONEY_OFF]);
		default:
			return(NULL);
	}
}



int eTypeCompareValues(script, type, v1, v2, res)
	ember	*script;
	int	type;
	e_value	*v1,
		*v2;
	int	*res;
{
	static	e_type	*typePtr;
	static	int	lastType = 0;

	if (type != lastType)
		typePtr = eTypeGetType(type);
	return((*typePtr->compare)(script,v1,v2,COND_EQ,res));
}


int eTypeGetBaseType(type)
	int	type;
{
	switch(type)
	{
		case EMBER_TYPE_INT:
			return(EMBER_TYPE_INT);
		case EMBER_TYPE_UINT:
			return(EMBER_TYPE_UINT);
		case EMBER_TYPE_REAL:
			return(EMBER_TYPE_REAL);
		case EMBER_TYPE_TEXT:
			return(EMBER_TYPE_TEXT);
		case EMBER_TYPE_MONEY:
			return(EMBER_TYPE_REAL);
		case EMBER_TYPE_ARRAY:
			return(EMBER_TYPE_ARRAY);
		case EMBER_TYPE_ASSOC:
			return(EMBER_TYPE_ASSOC);
	}
	fprintf(stderr,"\n\nFATAL : Unknown type %d\n\n",type);
	_debugTypeTrap();
	exit(1);
}


int eTypeGetRepType(type)
	int	type;
{
	switch(type)
	{
		case EMBER_TYPE_INT:
			return(EMBER_TYPE_INT);
		case EMBER_TYPE_UINT:
			return(EMBER_TYPE_UINT);
		case EMBER_TYPE_REAL:
			return(EMBER_TYPE_REAL);
		case EMBER_TYPE_TEXT:
			return(EMBER_TYPE_TEXT);
	}
	fprintf(stderr,"\n\nFATAL : Unknown type %d\n\n",type);
	_debugTypeTrap();
	exit(1);
}



int eTypeCheckMathOp(script, op, val1, val2)
	ember	*script;
	int	op;
	e_value	*val1,
		*val2;
{
	int	base1,
		base2;

	base1 = val1->type;
	base2 = val2->type;

	if (base1 == base2) return(0);
	if (base1 == EMBER_TYPE_UINT && base2 == EMBER_TYPE_INT) return(0);
	if (base1 == EMBER_TYPE_INT && base2 == EMBER_TYPE_UINT) return(0);
	if (base1 == EMBER_TYPE_REAL && base2 == EMBER_TYPE_INT) return(0);
	if (base1 == EMBER_TYPE_REAL && base2 == EMBER_TYPE_UINT) return(0);

	sprintf(script->errorMsg,"Bad types for maths operation (%s, %s)",
		val1->typePtr->name, val2->typePtr->name);
	return(-1);
}
